// WL_DRAW.C

#include "WL_DEF.H"
#include <DOS.H>
#pragma hdrstop


//#define DEBUGWALLS
//#define DEBUGTICS
/*
=============================================================================

						 LOCAL CONSTANTS

=============================================================================
*/

// the door is the last picture before the sprites
#define DOORWALL	(PMSpriteStart-8)

#define ACTORSIZE	0x4000

#define TRUCKTILE1      33 //Wall IDs for Truck tiles
#define TRUCKTILE2      63 
/*
=============================================================================

						 GLOBAL VARIABLES

=============================================================================
*/

long playx,playy;

boolean doarearain, playerisinside;

extern byte far shadetable[SHADE_COUNT][256];
extern int LSHADE_flag;

#ifdef DEBUGWALLS
unsigned screenloc[3]= {0,0,0};
#else
unsigned screenloc[3]= {PAGE1START,PAGE2START,PAGE3START};
#endif
unsigned freelatch = FREESTART;

long 	lasttimecount;
long 	frameon;
int     messagetime=0;
int     skippixel;

unsigned	        wallheight[MAXVIEWWIDTH], min_wallheight;

fixed	tileglobal	= TILEGLOBAL;
fixed	mindist		= MINDIST;


//
// math tables
//
int		pixelangle[MAXVIEWWIDTH];
long		far finetangent[FINEANGLES/4];
fixed 		far sintable[ANGLES+ANGLES/4],far *costable = sintable+(ANGLES/4);

//
// refresh variables
//
fixed	viewx,viewy;			// the focal point
int		viewangle;
fixed	viewsin,viewcos;



fixed	FixedByFrac (fixed a, fixed b);
void	TransformActor (objtype *ob);
void	BuildTables (void);
void    SetPixel (int x, int y, byte col);
void    ClearScreenH ();
void    PaintStars (long siny, long sinx);
void    PaintRain (long siny, long sinx);
void    PaintSnow (long siny, long sinx);
void	ClearScreen (void);
int	CalcRotate (objtype *ob);
void	DrawScaleds (void);
void	CalcTics (void);
void	FixOfs (void);
void	ThreeDRefresh (void);



//
// wall optimization variables
//
int		lastside;		// true for vertical
long	lastintercept;
int		lasttilehit;


//
// ray tracing variables
//
int		focaltx,focalty,viewtx,viewty;

int	       midangle,angle;
unsigned	xpartial,ypartial;
unsigned	xpartialup,xpartialdown,ypartialup,ypartialdown;
unsigned	xinttile,yinttile;

unsigned	tilehit;
unsigned	pixx;

int		xtile,ytile;
int		xtilestep,ytilestep;
long	        xintercept,yintercept;
long	        xstep,ystep;

int		horizwall[MAXWALLTILES],vertwall[MAXWALLTILES];


/*
=============================================================================

						 LOCAL VARIABLES

=============================================================================
*/


void AsmRefresh (void);			// in WL_DR_A.ASM

/*
============================================================================

			   3 - D  DEFINITIONS

============================================================================
*/


//==========================================================================


/*
========================
=
= FixedByFrac
=
= multiply a 16/16 bit, 2's complement fixed point number by a 16 bit
= fraction, passed as a signed magnitude 32 bit number
=
========================
*/

#pragma warn -rvl			// I stick the return value in with ASMs

fixed FixedByFrac (fixed a, fixed b)
{
//
// setup
//
asm	mov	si,[WORD PTR b+2]	// sign of result = sign of fraction

asm	mov	ax,[WORD PTR a]
asm	mov	cx,[WORD PTR a+2]

asm	or	cx,cx
asm	jns	aok:				// negative?
asm	neg	cx
asm	neg	ax
asm	sbb	cx,0
asm	xor	si,0x8000			// toggle sign of result
aok:

//
// multiply  cx:ax by bx
//
asm	mov	bx,[WORD PTR b]
asm	mul	bx					// fraction*fraction
asm	mov	di,dx				// di is low word of result
asm	mov	ax,cx				//
asm	mul	bx					// units*fraction
asm add	ax,di
asm	adc	dx,0

//
// put result dx:ax in 2's complement
//
asm	test	si,0x8000		// is the result negative?
asm	jz	ansok:
asm	neg	dx
asm	neg	ax
asm	sbb	dx,0

ansok:;

}

#pragma warn +rvl

//==========================================================================

/*
========================
=
= TransformActor
=
= Takes paramaters:
=   gx,gy		: globalx/globaly of point
=
= globals:
=   viewx,viewy		: point of view
=   viewcos,viewsin	: sin/cos of viewangle
=   scale		: conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
========================
*/


//
// transform actor
//
void TransformActor (objtype *ob)
{
	int ratio;
	fixed gx,gy,gxt,gyt,nx,ny;
	long	temp;

//
// translate point to view centered coordinates
//
	gx = ob->x-viewx;
	gy = ob->y-viewy;

//
// calculate newx
//
	gxt = FixedByFrac(gx,viewcos);
	gyt = FixedByFrac(gy,viewsin);
	nx = gxt-gyt-ACTORSIZE;		// fudge the shape forward a bit, because
								// the midpoint could put parts of the shape
								// into an adjacent wall

//
// calculate newy
//
	gxt = FixedByFrac(gx,viewsin);
	gyt = FixedByFrac(gy,viewcos);
	ny = gyt+gxt;

//
// calculate perspective ratio
//
	ob->transx = nx;
	ob->transy = ny;

	if (nx<mindist)			// too close, don't overflow the divide
	{
	  ob->viewheight = 0;
	  return;
	}

	ob->viewx = centerx + ny*scale/nx;	// DEBUG: use assembly divide

//
// calculate height (heightnumerator/(nx>>8))
//
	asm	mov	ax,[WORD PTR heightnumerator]
	asm	mov	dx,[WORD PTR heightnumerator+2]
	asm	idiv	[WORD PTR nx+1]			// nx>>8
	asm	mov	[WORD PTR temp],ax
	asm	mov	[WORD PTR temp+2],dx

	ob->viewheight = temp;
}

//==========================================================================

/*
========================
=
= TransformTile
=
= Takes paramaters:
=   tx,ty		: tile the object is centered in
=
= globals:
=   viewx,viewy		: point of view
=   viewcos,viewsin	: sin/cos of viewangle
=   scale		: conversion from global value to screen value
=
= sets:
=   screenx,transx,transy,screenheight: projected edge location and size
=
= Returns true if the tile is withing getting distance
=
========================
*/

boolean TransformTile (int tx, int ty, int *dispx, int *dispheight)
{
	int ratio;
	fixed gx,gy,gxt,gyt,nx,ny;
	long	temp;

//
// translate point to view centered coordinates
//
	gx = ((long)tx<<TILESHIFT)+0x8000-viewx;
	gy = ((long)ty<<TILESHIFT)+0x8000-viewy;

//
// calculate newx
//
	gxt = FixedByFrac(gx,viewcos);
	gyt = FixedByFrac(gy,viewsin);
	nx = gxt-gyt-0x2000;		// 0x2000 is size of object

//
// calculate newy
//
	gxt = FixedByFrac(gx,viewsin);
	gyt = FixedByFrac(gy,viewcos);
	ny = gyt+gxt;


//
// calculate perspective ratio
//
	if (nx<mindist)			// too close, don't overflow the divide
	{
		*dispheight = 0;
		return false;
	}

	*dispx = centerx + ny*scale/nx;	// DEBUG: use assembly divide

//
// calculate height (heightnumerator/(nx>>8))
//
	asm	mov	ax,[WORD PTR heightnumerator]
	asm	mov	dx,[WORD PTR heightnumerator+2]
	asm	idiv	[WORD PTR nx+1]			// nx>>8
	asm	mov	[WORD PTR temp],ax
	asm	mov	[WORD PTR temp+2],dx

	*dispheight = temp;

//
// see if it should be grabbed
//
	if (nx<TILEGLOBAL && ny>-TILEGLOBAL/2 && ny<TILEGLOBAL/2)
		return true;
	else
		return false;
}

//==========================================================================

/*
====================
=
= CalcHeight
=
= Calculates the height of xintercept,yintercept from viewx,viewy
=
====================
*/

#pragma warn -rvl			// I stick the return value in with ASMs

int	CalcHeight (void)
{
	int	transheight;
	int ratio;
	fixed gxt,gyt,nx,ny;
	long	gx,gy;

	gx = xintercept-viewx;
	gxt = FixedByFrac(gx,viewcos);

	gy = yintercept-viewy;
	gyt = FixedByFrac(gy,viewsin);

	nx = gxt-gyt;

  //
  // calculate perspective ratio (heightnumerator/(nx>>8))
  //
	if (nx<mindist)
		nx=mindist;			// don't let divide overflow

	asm	mov	ax,[WORD PTR heightnumerator]
	asm	mov	dx,[WORD PTR heightnumerator+2]
	asm	idiv	[WORD PTR nx+1]			// nx>>8
  //
  // Darkone's Textured ceiling/floor additional code
  //

	// update min_wallheight

	asm     mov     bx,[min_wallheight]
	asm     cmp     ax,bx
	asm     jae     noupdate:
	asm     mov     [min_wallheight],ax
	noupdate:;

	// End of additional code
}


//==========================================================================

/*
===================
=
= ScalePost
=
===================
*/

long		postsource;
unsigned	postx;
unsigned	postwidth;

void near ScalePost()
{
   byte far *vmem;
   int i,ywcount,yoffs,yw,yd,yendoffs;
   byte col,mask;

   byte shade;
   byte far *curshades;

   *((unsigned *)&vmem)=bufferofs;
   *(((unsigned *)&vmem)+1)=0xa000;

   for(i=0;i<postwidth;i++)
   {
      mask=1<<((postx+i)&3);
      VGAMAPMASK(mask);
      ywcount=yd=wallheight[postx+i]>>3;

      shade=(yd<<2)/((maxscaleshl2>>8)+1+LSHADE_flag);
      if(shade>32) shade=32; 
      else if(shade<1) shade=1;
      shade=32-shade;
      curshades=shadetable[shade];

      if(yd<=0) yd=100;
      yoffs=(viewheight/2-ywcount)*80;
      if(yoffs<0) yoffs=0;
      yoffs+=((postx+i)>>2);
      yendoffs=viewheight/2+ywcount-1;
      yw=63;
      while(yendoffs>=viewheight)
      {
         ywcount-=32;
         while(ywcount<=0)
         {
            ywcount+=yd;
            yw--;
         }
         yendoffs--;
      }
      if(yw<0) break;

      col=curshades[((byte far *)postsource)[yw]];

      yendoffs=yendoffs*80+((postx+i)>>2); 
      while(yoffs<=yendoffs) 
      { 
          vmem[yendoffs]=col; 

skippixel://<-Also Adams code 
          ywcount-=32; 
          if(ywcount<=0) 
          { 
             do 
             { 
                ywcount+=yd; 
                yw--; 
             } 
             while(ywcount<=0); 
             if(yw<0) break; 

	    col=curshades[((byte far *)postsource)[yw]];
	  }
	  yendoffs-=80;

          //Adams code:
          switch(MAPSPOT(xtile,ytile,0)) 
          {           
            case 64:                     // Outdoor textures. do not draw black 
            case 65:                     // pixels on the top half of the texture.

              shade=0;

              if(col==0 && yw<32)      
                    goto skippixel;  
                 else                  
                    break;
                                     
	  }
      }
      
   }
}

void  FarScalePost (void)				// just so other files can call
{
	ScalePost ();
}


/*
=======================================
=
= CheckHighBits
= BrotherTank's 64+ Wall values routine
=
=======================================
*/

int CheckHighBits (int value)                            // Check For Door Bit 7 only
  {
    if ((value & 0x40) && (value & 0x80))                // If Pushwall Return False Bit7&6
       { return 0; }
    else
       {
    if (value & 0x80)                                   // Is Door Return True
      { return 1; }
    else
      { return 0; }                                      // Not Door Return False
       }
  }

  int CheckAdjacentTile (int x, int y)
  {   // Part of 64+ Wall Tiles - Make sure within Map Boundaries
      // to fix bug of not allowing 64+ in outer edge of map
      // Checks Adjacent Tile for Bit 7 Set - Door or Pushwall
      if ((y-1 >=0) && (tilemap[x][y-1] & 0x80)
      || (y+1 <=63) && (tilemap[x][y+1] & 0x80)
      || (x-1 >=0) && (tilemap[x-1][y] & 0x80)
      || (x+1 <=63) && (tilemap[x+1][y] & 0x80))
	return 1;
      else
	return 0;
  }

/*
====================
=
= HitVertWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitVertWall (void)
{
	int		wallpic;
	unsigned	texture;

	texture = (yintercept>>4)&0xfc0;
        if(xtilestep == -1) 
        { 
           if(tilehit!=TRUCKTILE1 && tilehit!=TRUCKTILE2) 
              texture = 0xfc0-texture; 
              xintercept += TILEGLOBAL; 
        } 
	wallheight[pixx] = CalcHeight();

	// Fix for door graphic problem with 2 doors in a corner
	if (lastside==1 && lastintercept == xtile && lasttilehit == tilehit && !CheckAdjacentTile(xtile,ytile))

	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lastside = true;
		lastintercept = xtile;

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		if (CheckAdjacentTile(xtile, ytile))                            // 64+ Wall Graphics
		{								// check for adjacent doors
			ytile = yintercept>>TILESHIFT;
			if (CheckHighBits(tilemap[xtile-xtilestep][ytile]))
				wallpic = DOORWALL+3;
			else
				wallpic = vertwall[tilehit & ~0x40];
		}
		else
			wallpic = vertwall[tilehit];

		*(((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
                 (unsigned)postsource = texture;

	}
}


/*
====================
=
= HitHorizWall
=
= tilehit bit 7 is 0, because it's not a door tile
= if bit 6 is 1 and the adjacent tile is a door tile, use door side pic
=
====================
*/

void HitHorizWall (void)
{
	int		wallpic;
	unsigned	texture;

	texture = (xintercept>>4)&0xfc0;

        if(ytilestep == -1) 
           yintercept += TILEGLOBAL; 
        else if(tilehit!=TRUCKTILE1 && tilehit!=TRUCKTILE2) 
           texture = 0xfc0-texture;
 
	wallheight[pixx] = CalcHeight();

	// Fix for door graphic problem with 2 doors in a corner
	if (lastside==0 && lastintercept == ytile && lasttilehit == tilehit && !CheckAdjacentTile(xtile,ytile))
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lastside = 0;
		lastintercept = ytile;

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		if (CheckAdjacentTile(xtile,ytile))                             // 64+ Wall Graphics
		{								// check for adjacent doors
			xtile = xintercept>>TILESHIFT;
			if (CheckHighBits(tilemap[xtile][ytile-ytilestep]))
				wallpic = DOORWALL+2;
			else
				wallpic = horizwall[tilehit & ~0x40];
		}
		else
			wallpic = horizwall[tilehit];

		*(((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic); 
		 (unsigned)postsource = texture;
	}

}

//==========================================================================

/*
====================
=
= HitHorizDoor
=
====================
*/

void HitHorizDoor (void)
{
	unsigned	texture,doorpage,doornum;

	doornum = tilehit&0x7f;
	texture = ( (xintercept-doorposition[doornum]) >> 4) &0xfc0;

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
	// in the same door as last time, so check for optimized draw
		if (texture == (unsigned)postsource)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();			// draw last post
	// first pixel in this door
		lastside = 2;
		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		switch (doorobjlist[doornum].lock)
		{
		case dr_normal:
			doorpage = DOORWALL;
			break;
		case dr_lock1:
		case dr_lock2:
		case dr_lock3:
		case dr_lock4:
			doorpage = DOORWALL+6;
			break;
		case dr_elevator:
			doorpage = DOORWALL+4;
			break;
		}

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(doorpage);
		(unsigned)postsource = texture;
	}
}

//==========================================================================

/*
====================
=
= HitVertDoor
=
====================
*/

void HitVertDoor (void)
{
	unsigned	texture,doorpage,doornum;

	doornum = tilehit&0x7f;
	texture = ( (yintercept-doorposition[doornum]) >> 4) &0xfc0;

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
	// in the same door as last time, so check for optimized draw
		if (texture == (unsigned)postsource)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();			// draw last post
	// first pixel in this door
		lastside = 2;
		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		switch (doorobjlist[doornum].lock)
		{
		case dr_normal:
			doorpage = DOORWALL;
			break;
		case dr_lock1:
		case dr_lock2:
		case dr_lock3:
		case dr_lock4:
			doorpage = DOORWALL+6;
			break;
		case dr_elevator:
			doorpage = DOORWALL+4;
			break;
		}

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(doorpage+1);
		(unsigned)postsource = texture;
	}
}

//==========================================================================


/*
====================
=
= HitHorizPWall
=
= A pushable wall in action has been hit
=
====================
*/

void HitHorizPWall (void)
{
	int			wallpic;
	unsigned	texture,offset;

	texture = (xintercept>>4)&0xfc0;
	offset = pwallpos<<10;
	if (ytilestep == -1)
		yintercept += TILEGLOBAL-offset;
	else
	{
		texture = 0xfc0-texture;
		yintercept += offset;
	}

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		wallpic = horizwall[tilehit&63];

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
		(unsigned)postsource = texture;
	}

}

/*
====================
=
= HitVertPWall
=
= A pushable wall in action has been hit
=
====================
*/

void HitVertPWall (void)
{
	int			wallpic;
	unsigned	texture,offset;

	texture = (yintercept>>4)&0xfc0;
	offset = pwallpos<<10;
	if (xtilestep == -1)
	{
		xintercept += TILEGLOBAL-offset;
		texture = 0xfc0-texture;
	}
	else
		xintercept += offset;

	wallheight[pixx] = CalcHeight();

	if (lasttilehit == tilehit)
	{
		// in the same wall type as last time, so check for optimized draw
		if (texture == (unsigned)postsource)
		{
		// wide scale
			postwidth++;
			wallheight[pixx] = wallheight[pixx-1];
			return;
		}
		else
		{
			ScalePost ();
			(unsigned)postsource = texture;
			postwidth = 1;
			postx = pixx;
		}
	}
	else
	{
	// new wall
		if (lastside != -1)				// if not the first scaled post
			ScalePost ();

		lasttilehit = tilehit;
		postx = pixx;
		postwidth = 1;

		wallpic = vertwall[tilehit&63];

		*( ((unsigned *)&postsource)+1) = (unsigned)PM_GetPage(wallpic);
		(unsigned)postsource = texture;
	}

}

//==========================================================================

//==========================================================================

#if 0
/*
=============
=
= ClearScreen
=
=============
*/

void ClearScreen (void)
{
 unsigned floor=egaFloor[gamestate.episode*10+mapon],
	  ceiling=egaCeiling[gamestate.episode*10+mapon];

  //
  // clear the screen
  //
asm	mov	dx,GC_INDEX
asm	mov	ax,GC_MODE + 256*2		// read mode 0, write mode 2
asm	out	dx,ax
asm	mov	ax,GC_BITMASK + 255*256
asm	out	dx,ax

asm	mov	dx,40
asm	mov	ax,[viewwidth]
asm	shr	ax,3
asm	sub	dx,ax					// dx = 40-viewwidth/8

asm	mov	bx,[viewwidth]
asm	shr	bx,4					// bl = viewwidth/16
asm	mov	bh,BYTE PTR [viewheight]
asm	shr	bh,1					// half height

asm	mov	ax,[ceiling]
asm	mov	es,[screenseg]
asm	mov	di,[bufferofs]

toploop:
asm	mov	cl,bl
asm	rep	stosw
asm	add	di,dx
asm	dec	bh
asm	jnz	toploop

asm	mov	bh,BYTE PTR [viewheight]
asm	shr	bh,1					// half height
asm	mov	ax,[floor]

bottomloop:
asm	mov	cl,bl
asm	rep	stosw
asm	add	di,dx
asm	dec	bh
asm	jnz	bottomloop


asm	mov	dx,GC_INDEX
asm	mov	ax,GC_MODE + 256*10		// read mode 1, write mode 2
asm	out	dx,ax
asm	mov	al,GC_BITMASK
asm	out	dx,al

}
#endif
//==========================================================================
//
// Ripper's Outside Atmosphere
//
/*
==========
=
= SetPixel
=
==========
*/

void SetPixel(int x,int y,byte col)
{
   asm{
      cli
      mov al,2
      mov cl,byte ptr x
      and cl,3
      mov ah,1
      shl ah,cl
      mov dx,0x3c4
      out dx,ax
      sti
      push es
      push di
      mov ax,0xa000
      mov es,ax
      mov di,word ptr bufferofs
      mov bx,word ptr y
      mov ax,word ptr x
      shr ax,2
      add di,ax
      shl bx,4
      add di,bx
      shl bx,2
      add di,bx
      mov al,byte ptr col
      mov es:[di],al
      pop di
      pop es
   }
}

/*
==============
=
= ClearScreenH
=
==============
*/

void ClearScreenH()
{
   asm{
      cli
      mov al,2
      mov ah,15
      mov dx,0x3c4
      out dx,ax
      sti
      push es
      push di
      mov ax,0xa000
      mov es,ax
      mov di,word ptr bufferofs
      mov bx,320
      mov dx,word ptr viewwidth
      sub bx,dx
      shr dx,3
      shr bx,2
      mov dh,byte ptr viewheight
      shr dh,1
      xor ax,ax
      xor ch,ch
   }
nextrow:
   asm{
      mov cl,dl
      rep stosw
      add di,bx
      dec dh
      jne nextrow
      pop di
      pop es
   }
}

short far starsx[MAXSTARS];
short far starsy[MAXSTARS];
short far starsz[MAXSTARS];

byte moon[100]={
     0,  0, 27, 18, 15, 16, 19, 29,  0,  0,
     0, 22, 16, 15, 15, 16, 16, 18, 24,  0,
    27, 17, 15, 17, 16, 16, 17, 17, 18, 29,
    18, 15, 15, 15, 16, 16, 17, 17, 18, 20,
    16, 15, 15, 16, 16, 17, 17, 18, 19, 21,
    16, 15, 17, 20, 18, 17, 18, 18, 20, 22,
    19, 16, 18, 19, 17, 17, 18, 19, 22, 24,
    28, 19, 17, 17, 17, 18, 19, 21, 25, 31,
     0, 23, 18, 19, 18, 20, 22, 24, 28,  0,
     0,  0, 28, 21, 20, 22, 28, 30,  0,  0 };

#define LEVELSTATES 21

int levelSkyState[LEVELSTATES]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

/*
============
=
= PaintStars
=
============
*/

void PaintStars(long siny,long cosy)
{
   long x,y,z;
   int xx,yy;
   int i;
   int shade;
   int hvheight=viewheight>>1;
   int hvwidth=viewwidth>>1;
   ClearScreenH();
   for(i=0;i<MAXSTARS;i++)
   {
      x=(long)starsx[i]*cosy+(long)starsz[i]*siny;
      y=(long)starsy[i]<<16;
      z=((long)starsz[i]*cosy-(long)starsx[i]*siny)>>8;
      if(z<=0) continue;
      shade=(z>>18);
      if(shade>15) continue;
      xx=(int)(x/z+hvwidth);
      yy=(int)(hvheight-y/z);
      if(xx>=0 && xx<viewwidth && yy>=0 && yy<hvheight)
         SetPixel(xx,yy,shade+15);
   }

   x=16384L*cosy+16384L*siny;
   z=(16384L*cosy-16384L*siny)>>8;
   if(z<=0) return;
   xx=(int)(x/z+hvwidth);
   yy=(int)(hvheight-(((long)hvheight-((long)hvheight>>3))<<22)/z);
   if(xx>-10 && xx<viewwidth)
   {
      int stop=10,j;
      i=0;
      if(xx<0) i=-xx;
      if(xx>viewwidth-11) stop=viewwidth-xx;
      for(;i<stop;i++)
         for(j=0;j<10;j++)
            SetPixel(xx+i,yy+j,moon[j*10+i]);
   }
}

/*
===========
=
= PaintRain
=
===========
*/

word rainpos=0;

void PaintRain(long siny,long cosy)
{
   long px=(player->y+FixedByFrac(0x7900,siny))>>6;
   long pz=(player->x-FixedByFrac(0x7900,cosy))>>6;
   long ax,az;
   long x,y,z;
   int xx,yy,height,actheight;
   int i;
   int shade;
   int hvheight=viewheight>>1;
   int hvwidth=viewwidth>>1;

//--
   fixed    dist;
   boolean  flats;
   fixed    gu, gv;
   int      floorx,floory;
//--

   if(siny<0) siny=-(siny&0x7fffffff);
   if(cosy<0) cosy=-(cosy&0x7fffffff);

   rainpos-=1800;
   for(i=0;i<MAXSTARS;i++)
   {
      ax=(long)starsx[i]+px;
      ax=0x1fff-(ax&0x3fff);
      az=(long)starsz[i]+pz;
      az=0x1fff-(az&0x3fff);

      x=ax*cosy+az*siny;
      y=-(heightnumerator<<7)+((((long)(starsy[i]<<6)+(long)rainpos)&0x0ffff)<<11);
      z=(az*cosy-ax*siny)>>8;

      if(z<=0)
         continue;
      shade=(z>>17);
      if(shade>13)
          continue;

      xx=(int)(x/z+hvwidth);
      if(xx<0 || xx>viewwidth) 
         continue;

      actheight=(int)(y/z);
      yy=hvheight-actheight;

      height=(int)((heightnumerator<<10)/z);
      if(actheight<0) 
         actheight=-actheight;
      if(actheight<(wallheight[xx]>>3) && height<wallheight[xx]) 
         continue;

      if(xx>=0 && xx<viewwidth && yy>0 && yy<viewheight)
      {

         //Don't make rain inside.
         if(flats)
         {
            if (yy <= (hvheight - (wallheight[xx]>>3))) 
                     continue; 

            dist=((heightnumerator/((height>>3)+1))<<5); 
            gu= viewx+FixedByFrac(dist, viewcos); 
            gv=-viewy+FixedByFrac(dist, viewsin); 
            floorx = (  gu>>TILESHIFT   )&63; 
	    floory = (-(gv>>TILESHIFT)-1)&63;
 
            if(MAPSPOT(floorx,floory,2)>>8)
                     continue;

         }
         // the following lines can be optimized

         SetPixel(xx,yy,shade+15);
         SetPixel(xx,yy-1,shade+16);
         if(yy>2)
         {
            SetPixel(xx,yy-2,shade+17);
         }
      }
   }
}

/*
===========
=
= PaintSnow
=
===========
*/

void PaintSnow(long siny,long cosy)
{
   long px=(player->y+FixedByFrac(0x7900,siny))>>6;
   long pz=(player->x-FixedByFrac(0x7900,cosy))>>6;
   long ax,az;
   long x,y,z;
   int xx,yy,height,actheight;
   int i;
   int shade;
   int hvheight=viewheight>>1;
   int hvwidth=viewwidth>>1;

   if(siny<0) siny=-(siny&0x7fffffff);
   if(cosy<0) cosy=-(cosy&0x7fffffff);

   rainpos-=512;
   for(i=0;i<MAXSTARS;i++)
   {
      ax=(long)starsx[i]+px;
      ax=0x1fff-(ax&0x3fff);
      az=(long)starsz[i]+pz;
      az=0x1fff-(az&0x3fff);
      x=ax*cosy+az*siny;
      y=-(heightnumerator<<7)
        +((((long)(starsy[i]<<6)+(long)rainpos)&0x0ffff)<<11);
      z=(az*cosy-ax*siny)>>8;
      if(z<=0) continue;
      shade=(z>>17);
      if(shade>13) continue;
      xx=(int)(x/z+hvwidth);
      if(xx<0 || xx>viewwidth) continue;
      actheight=(int)(y/z);
      yy=hvheight-actheight;
      height=(int)((heightnumerator<<10)/z);
      if(actheight<0) actheight=-actheight;
      if(actheight<(wallheight[xx]>>3) && height<wallheight[xx]) continue;
      if(xx>0 && xx<viewwidth && yy>0 && yy<viewheight)
      {
         if(shade<10)
         {
            // the following lines can be optimized

            SetPixel(xx,yy,shade+17);
            SetPixel(xx-1,yy,shade+16);
            SetPixel(xx,yy-1,shade+16);
            SetPixel(xx-1,yy-1,shade+15);
         }
         else SetPixel(xx,yy,shade+15);
      }
   }
}
 



#define USE_TEX(page) (0x0000|(page))

/*
=======================================
=
= use USE_TEX(n) to use floor/ceiling
= textures for given level - n should
= be 1... 255 !!! (it is MAP-ID)
= 1st texture for floor, 2nd to ceiling
=
=======================================
*/

unsigned vgaCeiling[]=   // These aren't used anymore... Can be removed.
{
#ifndef SPEAR
 0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0xbfbf,
 0x4e4e,0x4e4e,0x4e4e,0x1d1d,0x8d8d,0x4e4e,0x1d1d,0x2d2d,0x1d1d,0x8d8d,
 0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x1d1d,0x2d2d,0xdddd,0x1d1d,0x1d1d,0x9898,

 0x1d1d,0x9d9d,0x2d2d,0xdddd,0xdddd,0x9d9d,0x2d2d,0x4d4d,0x1d1d,0xdddd,
 0x7d7d,0x1d1d,0x2d2d,0x2d2d,0xdddd,0xd7d7,0x1d1d,0x1d1d,0x1d1d,0x2d2d,
 0x1d1d,0x1d1d,0x1d1d,0x1d1d,0xdddd,0xdddd,0x7d7d,0xdddd,0xdddd,0xdddd
#else
 USE_TEX(74),USE_TEX(75),USE_TEX(76),USE_TEX(77),USE_TEX(78),USE_TEX(79),USE_TEX(80),USE_TEX(81),USE_TEX(82),
 USE_TEX(83),USE_TEX(84),USE_TEX(85),USE_TEX(86),USE_TEX(87),USE_TEX(88),USE_TEX(89),USE_TEX(90),USE_TEX(91),USE_TEX(92),USE_TEX(93)
#endif
};


// ------------------------- * Textured flats * -------------------------

/*
=============================
=
= Draw Textured Floor/Ceiling
=
=============================
*/

void DrawFlats(int tex_c, int tex_f)
{
   int x, y, y0, halfheight, shade;
   unsigned top_offset0, bot_offset0, top_offset, bot_offset;
   unsigned top_add, bot_add;
   byte p, color;
   byte far *src_bot, far *src_top;
   fixed dist;             // distance to row projection
   fixed tex_step;         // global step per one screen pixel
   fixed gu, gv, du, dv;   // global texture coordinates
   int u, v;               // local texture coordinates
   int floorx,floory;
   unsigned curfloortex, lastfloortex;

   // ------ * prepare * --------
   halfheight=viewheight>>1;
   y0=min_wallheight>>3;               // starting y value
   if(y0>halfheight) return;           // view obscued by walls
   if(y0==0) y0=1;                     // don't let division by zero
   top_offset0=80*(halfheight-y0-1);   // and will decrease by 80 each row
   bot_offset0=80*(halfheight+y0);     // and will increase by 80 each row
   //src_top=PM_GetPage(tex_c);        // multiple ceilings
   // draw horizontal lines

   lastfloortex=-1;
   for(p=0; p<4; p++)
   {
      asm mov ax,0x0102
      asm mov cl,[p]
      asm shl ah,cl
      asm mov dx,0x3c4
      asm out dx,ax

      for(y=y0, top_offset=top_offset0; y<halfheight; y++, top_offset-=80)
      {
	 dist=((heightnumerator/(y+1))<<5); 
	 gu= viewx+FixedByFrac(dist, viewcos);
	 gv=-viewy+FixedByFrac(dist, viewsin);
	 tex_step=(dist<<8)/viewwidth/175;
	 du= FixedByFrac(tex_step, viewsin);
	 dv=-FixedByFrac(tex_step, viewcos);
	 gu-=((viewwidth>>1)-p)*du;
	 gv-=((viewwidth>>1)-p)*dv; // starting point (leftmost)
	 du<<=2;                    // 4pix step
	 dv<<=2;

           shade=(y<<2)/((maxscaleshl2>>8)+1+LSHADE_flag);
	   if(shade>32) shade = 32;
	   else if(shade<1) shade = 1;
	   shade = 32-shade;

       
	 for(x=p, top_add=top_offset; x<viewwidth; x+=4, top_add++)
	 {
	    if(wallheight[x]>>3<=y)
	    {
	       floorx=(gu>>TILESHIFT)&63;
	       floory=(-(gv>>TILESHIFT)-1)&63;
	       curfloortex=MAPSPOT(floorx,floory,2)>>8; // ceiling

	       if(curfloortex != lastfloortex)
	       {
		  lastfloortex=curfloortex;
		  src_top=PM_GetPage(curfloortex);
	       }
	       u=(gu>>10)&63;
	       v=(gv>>10)&63;
	       color=shadetable[shade][*(src_top+((63-u)<<6)+(63-v))];

//Brothertanks:
//             if (color=0xff) // Transparency Color 
//---

	      // draw top pixel using <color>
	      if(curfloortex != 0)
	      {
	       asm mov es,[screenseg]
	       asm mov di,[bufferofs]
	       asm add di,[top_add]
	       asm mov al,[color]
	       asm mov es:[di],al
	      }
	    }
	    gu+=du;
	    gv+=dv;
	 }
      }
   }

   // src_bot=PM_GetPage(tex_f); // load floor texture
   lastfloortex=-1;
   for(p=0; p<4; p++)
   {
      asm mov ax,0x0102
      asm mov cl,[p]
      asm shl ah,cl
      asm mov dx,0x3c4
      asm out dx,ax

      for(y=y0, bot_offset=bot_offset0; y<halfheight; y++, bot_offset+=80)
      {
	 dist=((heightnumerator/(y+1))<<5); 
	 gu= viewx+FixedByFrac(dist, viewcos);
	 gv=-viewy+FixedByFrac(dist, viewsin);
	 tex_step=(dist<<8)/viewwidth/175;
	 du= FixedByFrac(tex_step, viewsin);
	 dv=-FixedByFrac(tex_step, viewcos);
	 gu-=((viewwidth>>1)-p)*du;
	 gv-=((viewwidth>>1)-p)*dv; // starting point (leftmost)
	 du<<=2;                    // 4pix step
	 dv<<=2;

	 shade=(y<<2)/((maxscaleshl2>>8)+1+LSHADE_flag);
	 if(shade>32) shade = 32;
	 else if(shade<1) shade = 1;
	 shade = 32-shade;

	 for(x=p, bot_add=bot_offset; x<viewwidth; x+=4, bot_add++)
	 {
	    if(wallheight[x]>>3<=y)
	    {
	       floorx = (gu>>TILESHIFT)&63;
	       floory = (-(gv>>TILESHIFT)-1)&63;
	       curfloortex=MAPSPOT(floorx,floory,2)&0xff;

	       if (curfloortex != lastfloortex)
	       {
		  lastfloortex=curfloortex;
		  src_bot=PM_GetPage(curfloortex);
	       }
	       u=(gu>>10)&63;
	       v=(gv>>10)&63;
	       color=shadetable[shade][*(src_bot+(u<<6)+(63-v))];

	       // draw bottom pixel using <color>

	       asm mov es,[screenseg]
	       asm mov di,[bufferofs]
	       asm add di,[bot_add]
	       asm mov al,[color]
	       asm mov es:[di],al
	    }
	    gu+=du;
	    gv+=dv;
	 }
      }
   }
}

/*
================
=
= VGAClearScreen
=
================
*/

void VGAClearScreen (void)//(unsigned ceiling)
{
 long siny,cosy;

 siny=(long)sintable[player->angle];
 cosy=(long)costable[player->angle];

   if(siny < 0) siny =-(siny&0x7fffffff);
   if(cosy < 0) cosy =-(cosy&0x7fffffff);

   PaintStars(siny,cosy);
// PaintRain(siny,cosy);
// PaintSnow(siny,cosy);
/*
   if(US_RndT()==1   || US_RndT()==80  ||  //Try pretend "Thunder" - needs some adjusting
      US_RndT()==160 || US_RndT()==240)

      StartBonusFlash();
   else{}
*/

}



//==========================================================================

/*
=====================
=
= CalcRotate
=
=====================
*/

int	CalcRotate (objtype *ob)
{
	int	angle,viewangle;

	// this isn't exactly correct, as it should vary by a trig value,
	// but it is close enough with only eight rotations

	viewangle = player->angle + (centerx - ob->viewx)/8;

	if (ob->obclass == rocketobj || ob->obclass == hrocketobj || ob->obclass == fireballobj)
		angle =  (viewangle-180)- ob->angle;
	else
		angle =  (viewangle-180)- dirangle[ob->dir];

	angle+=ANGLES/16;
	while (angle>=ANGLES)
		angle-=ANGLES;
	while (angle<0)
		angle+=ANGLES;

	if (ob->state->rotate == 2)             // 2 rotation pain frame
            return 0;//		return 4*(angle/(ANGLES/2));        // seperated by 3 (art layout...)

	return angle/(ANGLES/8);
}

/*
=====================
=
= DrawScaleds
=
= Draws all objects that are visable
=
=====================
*/

#define MAXVISABLE	50

typedef struct
{
	int	viewx,
		viewheight,
		shapenum;
	statobj_t *transsprite;
} visobj_t;

visobj_t	vislist[MAXVISABLE],*visptr,*visstep,*farthest;
void            Scale3DShape(statobj_t *ob);


void DrawScaleds (void)
{
	int 		i,j,least,numvisable,height;
	memptr		shape;
	byte		*tilespot,*visspot;
	int			shapenum;
	unsigned	spotloc;

	statobj_t	*statptr;
	objtype		*obj;

	visptr = &vislist[0];

//
// place static objects
//
	for (statptr = &statobjlist[0] ; statptr !=laststatobj ; statptr++)
	{
		if ((visptr->shapenum = statptr->shapenum) == -1)
			continue;						// object has been deleted

		if (!*statptr->visspot)
			continue;                                              // not visable
		//
		// WSJs animated sprites:
		//
		switch (statptr->shapenum)
		{
		case SPR_ANIM0:                      // Torch
		     visptr->shapenum+=(frameon>>2)%4;
		     break;
		case SPR_ANIM4:                      // Candle
#ifndef CONTEST
		     visptr->shapenum+=(frameon>>3)%2;
#else
		     visptr->shapenum+=(frameon>>7)%2;
#endif
		     break;
		case SPR_ANIM6:                      // Stove
		     visptr->shapenum+=(frameon>>2)%3;
		     break;
		case SPR_HANG_0:                     // Hanged doctor
		     visptr->shapenum+=(frameon>>2)%4;
		     break;
		case SPR_EMB_0:                      // Embryo
		     visptr->shapenum+=(frameon>>3)%3;
		     break;
		case SPR_DROP1:                      // Water drops
		     visptr->shapenum+=(frameon>>4)%4;
		     break;
		case SPR_TABLE01:                    // Hanging pot
		     visptr->shapenum+=(frameon>>2)%4;
		     break;
		case SPR_CANDLE01:                   // Candle hold
		     visptr->shapenum+=(frameon>>2)%4;
		     break;
		case SPR_FIRE01:                     // Campfire
		     visptr->shapenum+=(frameon>>3)%3;
		     break;
		case SPR_LIFE01:                     // Artefact of Life
		     visptr->shapenum+=(frameon>>5)%3;
		     break;
		case SPR_GOD01:                      // Artefact of Invulnerability
		     visptr->shapenum+=(frameon>>5)%3;
		     break;
		case SPR_BERSERK01:                  // Artefact of Fury
		     visptr->shapenum+=(frameon>>5)%3;
		     break;
		case SPR_STAT_0:                     // Artefact of Transportation
		     visptr->shapenum+=(frameon>>5)%3;
		     break;
		}

		if (TransformTile (statptr->tilex,statptr->tiley
			,&visptr->viewx,&visptr->viewheight) && statptr->flags & FL_BONUS)
		{
			GetBonus (statptr);
			continue;
		}

		if (!visptr->viewheight)
			continue;			// to close to the object
		if (statptr->flags & (FL_DIRSOUTH | FL_DIREAST))
		    visptr->transsprite=statptr;
		else
		    visptr->transsprite=NULL;
		if (visptr < &vislist[MAXVISABLE-1])	// don't let it overflow
			visptr++;
	}

//
// place active objects
//
	for (obj = player->next;obj;obj=obj->next)
	{
		if (!(visptr->shapenum = obj->state->shapenum))
			continue;						// no shape

		spotloc = (obj->tilex<<6)+obj->tiley;	// optimize: keep in struct?
		visspot = &spotvis[0][0]+spotloc;
		tilespot = &tilemap[0][0]+spotloc;

		//
		// could be in any of the nine surrounding tiles
		//
		if (*visspot
		|| ( *(visspot-1) && !*(tilespot-1) )
		|| ( *(visspot+1) && !*(tilespot+1) )
		|| ( *(visspot-65) && !*(tilespot-65) )
		|| ( *(visspot-64) && !*(tilespot-64) )
		|| ( *(visspot-63) && !*(tilespot-63) )
		|| ( *(visspot+65) && !*(tilespot+65) )
		|| ( *(visspot+64) && !*(tilespot+64) )
		|| ( *(visspot+63) && !*(tilespot+63) ) )
		{
			obj->active = true;
			TransformActor (obj);
			if (!obj->viewheight)
				continue;						// too close or far away

			visptr->viewx = obj->viewx;
			visptr->viewheight = obj->viewheight;
			if (visptr->shapenum == -1)
				visptr->shapenum = obj->temp1;	// special shape

			if (obj->state->rotate)
				visptr->shapenum += CalcRotate (obj);

			if (visptr < &vislist[MAXVISABLE-1])	// don't let it overflow

			{
				visptr->transsprite=NULL;

				visptr++;
			}
			obj->flags |= FL_VISABLE;
		}
		else
			obj->flags &= ~FL_VISABLE;
	}

//
// draw from back to front
//
	numvisable = visptr-&vislist[0];

	if (!numvisable)
		return;									// no visable objects

	for (i = 0; i<numvisable; i++)
	{
		least = 32000;
		for (visstep=&vislist[0] ; visstep<visptr ; visstep++)
		{
			height = visstep->viewheight;
			if (height < least)
			{
				least = height;
				farthest = visstep;
			}
		}
		//
		// draw farthest
		//       
		if(farthest->transsprite)
		   Scale3DShape(farthest->transsprite);
		else

		ScaleShape(farthest->viewx,farthest->shapenum,farthest->viewheight);

		farthest->viewheight = 32000;
	}

}

//==========================================================================

/*
==============
=
= DrawParallax
=
= by Ripper
=
==============
*/
void DrawParallax(int startpage)
{
   word xtex,nextxtex,offs;
   byte far *skytex;
   byte far *destbuf=(byte far *)(0xa0000000L+bufferofs);
   int i,x=0,x2,curtex,nextx=0,texoffs,t,tend;
   byte mask;
   int midangle=player->angle*(FINEANGLES/ANGLES);
   int skyheight=viewheight>>1;
   startpage+=15;

   t=pixelangle[0]+midangle;
   while(t<0) t+=FINEANGLES;
   while(t>=FINEANGLES) t-=FINEANGLES;
   xtex=(word)((((long)t)<<10)/FINEANGLES);

   do
   {
      curtex=xtex>>6;
      skytex=(byte far *)(((long)(unsigned) PM_GetPage(startpage-curtex))<<16);
      nextx=0x7fff;
      for(i=0;i<4;i++)
      {
         mask=1<<((x+i)&3);
         VGAMAPMASK(mask);
         for(x2=x+i;x2<viewwidth;x2+=4)
         {
            t=pixelangle[x2]+midangle;
            while(t<0) t+=FINEANGLES;
            while(t>=FINEANGLES) t-=FINEANGLES;
            xtex=(word)((((long)t)<<10)/FINEANGLES);

            t=xtex>>6;
            if(t!=curtex)
            {
               if(x2<nextx) nextx=x2,nextxtex=xtex;
               break;
            }

            texoffs=0xfc0-((xtex&63)<<6);
            tend=skyheight-(wallheight[x2]>>3);
            if(tend<=0) continue;
            for(t=0,offs=x2>>2;t<tend;t++,offs+=80)
               destbuf[offs]=skytex[texoffs+((t<<6)/skyheight)];
         }
      }
      x=nextx;
      xtex=nextxtex;
   }
   while(x<viewwidth);
}

/*
=========================
=
= DrawPlayerWeapon
=
= Draw the player's hands
=
=========================
*/

int	weaponscale[NUMWEAPONS] = {SPR_KNIFEREADY,SPR_PISTOLREADY
	,SPR_MACHINEGUNREADY,SPR_CHAINREADY,SPR_STENREADY,SPR_SPECTRE_W1
	,SPR_SPECTRE_F1};

void DrawPlayerWeapon (void)
{
	int	shapenum;

#ifndef SPEAR
	if (gamestate.victoryflag)
	{
		if (player->state == &s_deathcam && (TimeCount&32) )
			SimpleScaleShape(viewwidth/2,SPR_DEATHCAM,viewheight+1);
		return;
	}
#endif

	if (gamestate.weapon != -1)
	{
		shapenum = weaponscale[gamestate.weapon]+gamestate.weaponframe;
		SimpleScaleShape(viewwidth/2-10+gamestate.bobber2,shapenum,viewheight+1+gamestate.bobber+gamestate.weaponchange);
	}

	if (demorecord || demoplayback)
		SimpleScaleShape(viewwidth/2,SPR_DEMO,viewheight+1);
}


//==========================================================================


/*
=====================
=
= CalcTics
=
=====================
*/

void CalcTics (void)
{
	long	newtime,oldtimecount;

//
// calculate tics since last refresh for adaptive timing
//
	if (lasttimecount > TimeCount)
		TimeCount = lasttimecount;		// if the game was paused a LONG time

	do
	{
		newtime = TimeCount;
		tics = newtime-lasttimecount;
	} while (!tics);			// make sure at least one tic passes

	lasttimecount = newtime;

#ifdef FILEPROFILE
		strcpy (scratch,"\tTics:");
		itoa (tics,str,10);
		strcat (scratch,str);
		strcat (scratch,"\n");
		write (profilehandle,scratch,strlen(scratch));
#endif

	if (tics>MAXTICS)
	{
		TimeCount -= (tics-MAXTICS);
		tics = MAXTICS;
	}
}


//==========================================================================


/*
========================
=
= FixOfs
=
========================
*/

void	FixOfs (void)
{
	VW_ScreenToScreen (displayofs,bufferofs,viewwidth/8,viewheight);
}


//==========================================================================


/*
====================
=
= WallRefresh
=
====================
*/

void WallRefresh (void)
{
//
// set up variables for this view
//
	viewangle = player->angle;
	midangle = viewangle*(FINEANGLES/ANGLES);
	viewsin = sintable[viewangle];
	viewcos = costable[viewangle];
	viewx = player->x - FixedByFrac(focallength,viewcos);
	viewy = player->y + FixedByFrac(focallength,viewsin);

	focaltx = viewx>>TILESHIFT;
	focalty = viewy>>TILESHIFT;

	viewtx = player->x >> TILESHIFT;
	viewty = player->y >> TILESHIFT;

	xpartialdown = viewx&(TILEGLOBAL-1);
	xpartialup = TILEGLOBAL-xpartialdown;
	ypartialdown = viewy&(TILEGLOBAL-1);
	ypartialup = TILEGLOBAL-ypartialdown;

	lastside = -1;			// the first pixel is on a new wall
	AsmRefresh ();
	ScalePost ();			// no more optimization on last post
}

//==========================================================================

/*
==============
=
= ThreeDRefresh
=
===============
*/

void	ThreeDRefresh (void)
{
	int tracedir;
	unsigned ceiling;
	boolean flats;

// this wouldn't need to be done except for my debugger/video wierdness
	outportb (SC_INDEX,SC_MAPMASK);

//
// clear out the traced array
//
asm	mov	ax,ds
asm	mov	es,ax
asm	mov	di,OFFSET spotvis
asm	xor	ax,ax
asm	mov	cx,2048	      // 64*64 / 2
asm	rep stosw

	bufferofs += screenofs;

//
// follow the walls from there to the right, drawwing as we go
//
   ceiling=vgaCeiling[gamestate.episode*10+mapon];
   if((ceiling>>8)==(ceiling&0xfF))
   {
	VGAClearScreen();

	flats=false;
   }
   else
   {
	flats=true;
	ceiling=((ceiling&0xFF)-1)<<1;
   }

	if(levelSkyState) VGAClearScreen(); // Draw the sky
			
	WallRefresh ();

	if(mapsegs[0][63*64]) DrawParallax(mapsegs[0][63*64]);

	if(flats)
	   DrawFlats(ceiling, ceiling+1); // draw textured floor/ceiling

//
// draw all the scaled images
//
	playy=(player->y+FixedByFrac(0x7300,viewsin));
	playx=(player->x-FixedByFrac(0x7300,viewcos));


	DrawScaleds();			// draw scaled stuff
//      PaintRain(viewsin,viewcos);     // draw rain - no rain today.

//	tracedir=gamestate.episode*10+mapon;
//	if(tracedir>=LEVELSTATES) tracedir=LEVELSTATES-1;
							   // For Ripper Outside atmosphere... Unused
//	if(levelSkyState[tracedir]==3)
//	   PaintSnow(viewsin,viewcos);
//	else if(levelSkyState[tracedir]==2)
//	   PaintRain(viewsin,viewcos);


	DrawPlayerWeapon ();     	// draw player's hands
	if (messagetime > 0)            // draw message
	   DrawMessage ();

//
// show screen and time last cycle
//
	if (fizzlein)
	{
		FizzleFade(bufferofs,displayofs+screenofs,viewwidth,viewheight,20,false);
		fizzlein = false;

		lasttimecount = TimeCount = 0;		// don't make a big tic count

	}

        //
        // Define Flamer46's Timer
        //
        gamestate.finaltime -= tics; 
        if(gamestate.finaltime>0) 
                     BonusTimer(); //If finaltime greater than 0, call the clock 


	bufferofs -= screenofs;
	displayofs = bufferofs;

	asm	cli
	asm	mov	cx,[displayofs]
	asm	mov	dx,3d4h		// CRTC address register
	asm	mov	al,0ch		// start address high register
	asm	out	dx,al
	asm	inc	dx
	asm	mov	al,ch
	asm	out	dx,al   	// set the high byte
	asm	sti

	bufferofs += SCREENSIZE;
	if (bufferofs > PAGE3START)
		bufferofs = PAGE1START;

	frameon++;
	PM_NextFrame();
}

/*
=================================
=
= BonusTimer display
= by InsurrectionMan and Flamer46
=
================================
*/

void BonusTimer() 
{ 
   int min,sec; 

   fontnumber=0;
   if(gamestate.finaltime>0)
      sec=gamestate.finaltime/70;
   else
      sec%=60;

   min=sec/60; 
  

   SETFONTCOLOR(0xaf,0x5f); //Timer color 
   PrintX=270; PrintY=2;    //Timer location
 
   US_PrintUnsigned(min/10); 
   US_PrintUnsigned(min%10);
   US_Print(":"); 
   US_PrintUnsigned(sec/10);
   US_PrintUnsigned(sec%10);
} 

/*
=======================================
=
=GetMessage
=
=gets ingame messages
=
=======================================
*/

void GetMessage (char *lastmessage)
{
   messagetime = 150; // time fo message to display

   strcpy(gamestate.message, lastmessage);
}

/*
============================
=
=DrawMessages
=
=displays ingame messages
=
============================
*/

void DrawMessage (void)
{
   if(messagetime <= 0) { strcpy (gamestate.message,""); return; }
   messagetime-=tics;

   fontnumber = 0;
   SETFONTCOLOR(0xaf,0x5f);                 // set the color (red, black)
   PrintX=8; PrintY=2;                      // position the message
   US_Print(gamestate.message);             // print message
   DrawAllPlayBorderSides ();
}
//===========================================================================

